"""CentriConnect/MyPropane API class"""

import asyncio
import json
import socket
from json.decoder import JSONDecodeError

import aiohttp
import async_timeout
import backoff

from .exceptions import (
    CentriConnectConnectionError,
    CentriConnectConnectionTimeoutError,
    CentriConnectDecodeError,
    CentriConnectEmptyResponseError,
    CentriConnectNotFoundError,
    CentriConnectTooManyRequestsError,
)
from .types import TankDict


class API:
    """This class reads data from the CentriConnect/MyPropane API."""

    def __init__(
        self,
        user_id: str,
        device_id: str,
        device_auth: str,
        session: aiohttp.ClientSession,
        timeout: float = 2,
    ):
        self.session = session
        self.timeout = timeout
        self.user_id = user_id
        self.device_id = device_id
        self.device_auth = device_auth

    @backoff.on_exception(
        backoff.expo,
        (CentriConnectConnectionTimeoutError, CentriConnectConnectionError),
        max_tries=3,
    )
    async def async_request(self) -> TankDict:
        """Make an asynchronous request to the CentriConnect/MyPropane API.
        Raises:
            CentriConnectConnectionTimeoutError: The connection timed out
            CentriConnectConnectionError: Any communication error
            CentriConnectNotFoundError: The requested device was not found
            CentriConnectTooManyRequestsError: Too many requests were made to the API
            CentriConnectDecodeError: The API response could not be decoded
            CentriConnectEmptyResponseError: The API response did not contain any data

        Returns:
            Typed dictionary containing the tank data
        """
        try:
            async with async_timeout.timeout(self.timeout):
                async with self.session.get(self.build_url()) as response:
                    return await self._handle_response(response)
        except asyncio.TimeoutError as ex:
            raise CentriConnectConnectionTimeoutError(
                "Timeout while connecting to CentriConnect API"
            ) from ex
        except (aiohttp.ClientError, socket.gaierror) as ex:
            raise CentriConnectConnectionError(
                f"Error while communicating with CentriConnect API: {ex}"
            ) from ex

    def build_url(self) -> str:
        """Construct the URL for a specified user and device."""
        return (
            f"https://api.centriconnect.com/centriconnect/{self.user_id}"
            + f"/device/{self.device_id}/all-data"
            + f"?device_auth={self.device_auth}"
        )

    async def _handle_response(self, response: aiohttp.ClientResponse) -> TankDict:
        """Handle the API response."""
        if response.status == 404:
            raise CentriConnectNotFoundError()
        if response.status == 429:
            raise CentriConnectTooManyRequestsError()
        if not response.ok:
            raise CentriConnectConnectionError(
                f"Received unexpected status code {response.status} from CentriConnect API"
            )
        raw_body = await response.text()
        try:
            json_body = json.loads(raw_body)
            if json_body is None:
                raise CentriConnectEmptyResponseError(
                    "CentriConnect API response did not contain expected data", raw_body
                )
            if "error" in json_body:
                error_msg = json_body["error"]
                if error_msg == "NotFound":
                    raise CentriConnectNotFoundError()
                else:
                    raise CentriConnectDecodeError(
                        "CentriConnect API response did not contain expected data",
                        raw_body,
                    )
            if self.device_id not in json_body:
                raise CentriConnectDecodeError(
                    "CentriConnect API response did not contain expected data",
                    raw_body,
                )
            return json_body[self.device_id]
        except JSONDecodeError as ex:
            raise CentriConnectDecodeError(
                "Error decoding JSON response from CentriConnect", raw_body
            ) from ex



"""Data class for CentriConnect/MyPropane Tank information."""

from datetime import datetime

from aiocentriconnect.types import TankDict


class Tank:
    """Object holding information about a CentriConnect/MyPropane tank."""

    def __init__(self, raw_data: TankDict):
        self.raw_data = raw_data

    @property
    def alert_status(self) -> str:
        """Alert status"""
        return self.raw_data["AlertStatus"]

    @property
    def altitude(self) -> float:
        """Altitude"""
        return self.raw_data["Altitude"]

    @property
    def battery_level(self) -> float:
        """Battery level"""
        # The battery level is estimated based on the battery voltage,
        # with 3.5V or below being 0% and 4.05V or above being 100%.
        return min(1.0, max(((self.battery_voltage - 3.5) / 0.5), 0.0)) * 100

    @property
    def battery_voltage(self) -> float:
        """Battery voltage"""
        return self.raw_data["BatteryVolts"]

    @property
    def device_id(self) -> str:
        """Device ID"""
        return self.raw_data["DeviceID"]

    @property
    def device_name(self) -> str:
        """Device name"""
        return self.raw_data["DeviceName"]

    @property
    def device_temperature(self) -> float:
        """Device temperature"""
        return self.raw_data["DeviceTempFahrenheit"]

    @property
    def last_post_time(self) -> datetime:
        """Last post time"""
        string_time = self.raw_data["LastPostTimeIso"]
        date = string_time.replace(" ", "T") + "Z"
        return datetime.fromisoformat(date)

    @property
    def latitude(self) -> float:
        """Latitude"""
        return self.raw_data["Latitude"]

    @property
    def longitude(self) -> float:
        """Longitude"""
        return self.raw_data["Longitude"]

    @property
    def lte_signal_level(self) -> float:
        """LTE signal level"""
        # The LTE signal level is estimated based on the LTE signal strength,
        # with -140 dBm or below being 0% and -70 dBm or above being 100%.
        return min(1.0, max(((self.lte_signal_strength + 140.0) / 70.0), 0.0)) * 100

    @property
    def lte_signal_strength(self) -> float:
        """LTE signal strength"""
        return self.raw_data["SignalQualLTE"]

    @property
    def next_post_time(self) -> datetime:
        """Next post time"""
        string_time = self.raw_data["NextPostTimeIso"]
        date = string_time.replace(" ", "T") + "Z"
        return datetime.fromisoformat(date)

    @property
    def solar_level(self) -> float:
        """Solar level"""
        # The solar level is estimated based on the solar voltage,
        # with 0V being 0% and 2.6V or above being 100%.
        return min(1.0, max((self.solar_voltage / 2.6), 0.0)) * 100

    @property
    def solar_voltage(self) -> float:
        """Solar voltage"""
        return self.raw_data["SolarVolts"]

    @property
    def tank_level(self) -> float:
        """Tank level"""
        return self.raw_data["TankLevel"]

    @property
    def tank_remaining_volume(self) -> float:
        """Tank remaining volume"""
        return self.tank_level * 0.01 * self.tank_size

    @property
    def tank_size(self) -> int:
        """Tank size"""
        return self.raw_data["TankSize"]

    @property
    def tank_size_unit(self) -> str:
        """Tank size unit"""
        return self.raw_data["TankSizeUnit"]

    @property
    def hardware_version(self) -> str:
        """Hardware version"""
        return self.raw_data["VersionHW"]

    @property
    def lte_version(self) -> str:
        """LTE version"""
        return self.raw_data["VersionLTE"]



"""Tests for CentriConnect/MyPropane API"""
# pylint: disable=missing-function-docstring

import asyncio

import aiohttp
from aiohttp.web import Response
import pytest

from aiocentriconnect import CentriConnect
from aiocentriconnect.exceptions import (
    CentriConnectConnectionError,
    CentriConnectConnectionTimeoutError,
    CentriConnectDecodeError,
    CentriConnectNotFoundError,
    CentriConnectTooManyRequestsError,
)
from aresponses.main import ResponsesMockServer  # type: ignore

fake_user_id = "12345678-9012-3456-7a89-b012345cde6f"
fake_device_id = "123a4b5c-678d-9e0f-a123-4b567c8d901e"
fake_device_auth = "123456"
fake_response = """
{
    "123a4b5c-678d-9e0f-a123-4b567c8d901e": {
        "AlertStatus": "No Alert",
        "Altitude": 123.456,
        "BatteryVolts": 4.19,
        "DeviceID": "123a4b5c-678d-9e0f-a123-4b567c8d901e",
        "DeviceName": "My Tank",
        "DeviceTempCelsius": 17.0,
        "DeviceTempFahrenheit": 63.0,
        "LastPostTimeIso": "2026-02-27 22:00:31.000",
        "Latitude": 40.7128,
        "Longitude": -74.0060,
        "NextPostTimeIso": "2026-02-28 10:00:00.000",
        "SignalQualLTE": -107.0,
        "SolarVolts": 2.46,
        "TankLevel": 75.0,
        "TankSize": 1000,
        "TankSizeUnit": "Gallons",
        "VersionHW": "4.1",
        "VersionLTE": "1.1.2"
    }
}
"""
api_host = "api.centriconnect.com"
api_path = (
    f"/centriconnect/{fake_user_id}/device/{fake_device_id}"
    + f"/all-data?device_auth={fake_device_auth}"
)


@pytest.mark.asyncio
async def test_tank_data_request(aresponses: ResponsesMockServer):
    add_valid_tank_data_response(aresponses)

    async with aiohttp.ClientSession() as session:
        centriconnect = CentriConnect(
            fake_user_id, fake_device_id, fake_device_auth, session=session
        )
        tank_data = await centriconnect.async_get_tank_data()
        assert tank_data.alert_status == "No Alert"
        assert tank_data.altitude == 123.456
        assert tank_data.battery_level == 100.0
        assert tank_data.battery_voltage == 4.19
        assert tank_data.device_id == fake_device_id
        assert tank_data.device_name == "My Tank"
        assert tank_data.device_temperature == 63.0
        assert tank_data.latitude == 40.7128
        assert tank_data.longitude == -74.0060
        assert tank_data.last_post_time.isoformat() == "2026-02-27T22:00:31+00:00"
        assert tank_data.next_post_time.isoformat() == "2026-02-28T10:00:00+00:00"
        assert tank_data.lte_signal_level == 47.14285714285714
        assert tank_data.lte_signal_strength == -107.0
        assert tank_data.solar_level == 94.61538461538461
        assert tank_data.solar_voltage == 2.46
        assert tank_data.tank_level == 75.0
        assert tank_data.tank_remaining_volume == 750.0
        assert tank_data.tank_size == 1000.0
        assert tank_data.tank_size_unit == "Gallons"
        assert tank_data.hardware_version == "4.1"
        assert tank_data.lte_version == "1.1.2"

    aresponses.assert_plan_strictly_followed()


@pytest.mark.asyncio
async def test_internal_session(aresponses: ResponsesMockServer):
    add_valid_tank_data_response(aresponses)

    async with CentriConnect(
        fake_user_id, fake_device_id, fake_device_auth
    ) as centriconnect:
        tank_data = await centriconnect.async_get_tank_data()
        assert tank_data.alert_status == "No Alert"

    aresponses.assert_plan_strictly_followed()


@pytest.mark.asyncio
async def test_timeout(aresponses: ResponsesMockServer):
    async def response_handler(_):
        await asyncio.sleep(0.2)
        return get_valid_tank_data_response(aresponses)

    # Backoff will try 3 times
    for _ in range(3):
        aresponses.add(  # type: ignore
            api_host,  # type: ignore
            api_path,  # type: ignore
            "GET",  # type: ignore
            response_handler,  # type: ignore
            match_querystring=True,
        )

    async with aiohttp.ClientSession() as session:
        centriconnect = CentriConnect(
            fake_user_id, fake_device_id, fake_device_auth, timeout=0.1, session=session
        )
        with pytest.raises(CentriConnectConnectionTimeoutError):
            assert await centriconnect.async_get_tank_data()

    aresponses.assert_plan_strictly_followed()


@pytest.mark.asyncio
async def test_weird_error_response(aresponses: ResponsesMockServer):
    aresponses.add(  # type: ignore
        api_host,  # type: ignore
        api_path,  # type: ignore
        "GET",  # type: ignore
        aresponses.Response(
            status=500, headers={"Content-Type": "text/plain"}, text="Error"
        ),  # type: ignore
        match_querystring=True,
    )

    async with aiohttp.ClientSession() as session:
        centriconnect = CentriConnect(
            fake_user_id, fake_device_id, fake_device_auth, session=session
        )
        with pytest.raises(CentriConnectConnectionError):
            assert await centriconnect.async_get_tank_data()


@pytest.mark.asyncio
async def test_not_found_error_response(aresponses: ResponsesMockServer):
    aresponses.add(  # type: ignore
        api_host,  # type: ignore
        api_path,  # type: ignore
        "GET",  # type: ignore
        aresponses.Response(
            status=404,
            headers={"Content-Type": "text/plain"},
            text='{"error": "NotFound", "message": "Device does not exist."}',
        ),  # type: ignore
        match_querystring=True,
    )

    async with aiohttp.ClientSession() as session:
        centriconnect = CentriConnect(
            fake_user_id, fake_device_id, fake_device_auth, session=session
        )
        with pytest.raises(CentriConnectNotFoundError):
            assert await centriconnect.async_get_tank_data()


@pytest.mark.asyncio
async def test_not_too_many_requests_response(aresponses: ResponsesMockServer):
    aresponses.add(  # type: ignore
        api_host,  # type: ignore
        api_path,  # type: ignore
        "GET",  # type: ignore
        aresponses.Response(
            status=429,
            headers={"Content-Type": "text/plain"},
            text="",
        ),  # type: ignore
        match_querystring=True,
    )

    async with aiohttp.ClientSession() as session:
        centriconnect = CentriConnect(
            fake_user_id, fake_device_id, fake_device_auth, session=session
        )
        with pytest.raises(CentriConnectTooManyRequestsError):
            assert await centriconnect.async_get_tank_data()


@pytest.mark.asyncio
async def test_can_get_raw_response_from_exception(aresponses: ResponsesMockServer):
    malformed_json = """
    {
        "123a4b5c-678d-9e0f-a123-4b567c8d901e": {
            "AlertStatus": "No Alert",
            "Altitude": 123.456,
            "BatteryVolts": 4.19,
    """
    aresponses.add(  # type: ignore
        api_host,  # type: ignore
        api_path,  # type: ignore
        "GET",  # type: ignore
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=malformed_json,
        ),  # type: ignore
        match_querystring=True,
    )

    async with aiohttp.ClientSession() as session:
        centriconnect = CentriConnect(
            fake_user_id, fake_device_id, fake_device_auth, session=session
        )
        try:
            await centriconnect.async_get_tank_data()
        except CentriConnectDecodeError as ex:
            assert ex.get_raw_body() == malformed_json

    aresponses.assert_plan_strictly_followed()


def add_valid_tank_data_response(aresponses: ResponsesMockServer) -> None:
    aresponses.add(  # type: ignore
        api_host,  # type: ignore
        api_path,  # type: ignore
        "GET",  # type: ignore
        get_valid_tank_data_response(aresponses),  # type: ignore
        match_querystring=True,
    )


def get_valid_tank_data_response(
    aresponses: ResponsesMockServer,
) -> Response:
    return aresponses.Response(
        status=200,
        headers={"Content-Type": "application/json"},
        text=fake_response,
    )



"""Asynchronous Python client for CentriConnect/MyPropane API."""

from typing import Optional

from aiohttp import ClientSession

from .api import API
from .tank import Tank


class CentriConnect:
    """Main class for reading data for a CentriConnect/MyPropane tank."""

    def __init__(
        self,
        user_id: str,
        device_id: str,
        device_auth: str,
        session: Optional[ClientSession] = None,
        timeout: float = 1,
    ):
        if session is None:
            session = ClientSession()
            self._session = session
        self.api = API(user_id, device_id, device_auth, session, timeout)

    async def async_get_tank_data(self) -> Tank:
        """Get tank data"""
        return Tank(await self.api.async_request())

    async def __aenter__(self):
        return self

    async def __aexit__(self, *_exc_info) -> None:  # pyright: ignore[reportUnknownParameterType, reportMissingParameterType]
        if self._session:
            await self._session.close()



"""CentriConnect Exceptions"""


class CentriConnectError(Exception):
    """Generic CentriConnect exception."""


class CentriConnectDecodeError(CentriConnectError):
    """CentriConnect exception thrown when decoding a response fails."""

    def __init__(self, message: str, raw_body: str):
        """Init Decode error."""
        super().__init__(message)
        self.raw_body = raw_body

    def get_raw_body(self) -> str:
        """Return the raw body of the failing request."""
        return self.raw_body


class CentriConnectEmptyResponseError(CentriConnectError):
    """CentriConnect empty API response exception."""


class CentriConnectNotFoundError(CentriConnectError):
    """CentriConnect device not found exception."""


class CentriConnectTooManyRequestsError(CentriConnectError):
    """CentriConnect too many requests exception."""


class CentriConnectConnectionError(CentriConnectError):
    """CentriConnect connection exception."""


class CentriConnectConnectionTimeoutError(CentriConnectConnectionError):
    """CentriConnect connection timeout exception."""



""" Simple example showing usage of this library """
# pylint: disable=missing-function-docstring

import asyncio

from aiocentriconnect import CentriConnect


async def main():
    user_id = input("User ID: ")
    device_id = input("Device ID: ")
    device_auth = input("Device auth code: ")

    async with CentriConnect(user_id, device_id, device_auth) as centriconnect:
        tank_data = await centriconnect.async_get_tank_data()
        print(f"'{tank_data.device_name}' is {tank_data.tank_level}% full")


asyncio.run(main())



"""Asynchronous Python client for CentriConnect/MyPropane API."""

from .centriconnect import CentriConnect  # type: ignore # noqa
from .tank import Tank  # type: ignore # noqa

__version__ = "0.2.3"



"""Inner types for CentriConnect/MyPropane API."""

from typing import TypedDict


class TankDict(TypedDict):
    """Typed dictionary for a CentriConnect/MyPropane tank."""

    AlertStatus: str
    Altitude: float
    BatteryVolts: float
    DeviceID: str
    DeviceName: str
    DeviceTempFahrenheit: float
    LastPostTimeIso: str
    Latitude: float
    Longitude: float
    NextPostTimeIso: str
    SignalQualLTE: float
    SolarVolts: float
    TankLevel: float
    TankSize: int
    TankSizeUnit: str
    VersionHW: str
    VersionLTE: str
